@@ -25,3 +25,6 @@ class PortfolioPlace(Handy.ActionRow):
     eject = Gtk.Template.Child()
     mount = None
     path = ""
+    uuid = ""
+    encrypted = None
+    device = None
