@@ -18,7 +18,8 @@
 import os
 
 from pathlib import Path
-from locale import gettext as _
+
+from .translation import gettext as _
 
 from gi.repository import Gtk, GLib, Gio, Handy, GObject
 
@@ -31,7 +32,13 @@ from .worker import PortfolioDeleteWorker
 from .worker import PortfolioLoadWorker
 from .worker import PortfolioOpenWorker
 from .worker import PortfolioPropertiesWorker
+from .worker import PortfolioRestoreTrashWorker
+from .worker import PortfolioDeleteTrashWorker
+from .worker import PortfolioSendTrashWorker
+from .worker import PortfolioLoadTrashWorker
 from .places import PortfolioPlaces
+from .settings import PortfolioSettings
+from .trash import default_trash
 
 
 @Gtk.Template(resource_path="/dev/tchx84/Portfolio/window.ui")
@@ -57,15 +64,19 @@ class PortfolioWindow(Handy.ApplicationWindow):
     select_all = Gtk.Template.Child()
     select_none = Gtk.Template.Child()
     new_folder = Gtk.Template.Child()
+    delete_trash = Gtk.Template.Child()
+    restore_trash = Gtk.Template.Child()
     loading_label = Gtk.Template.Child()
     loading_bar = Gtk.Template.Child()
     loading_description = Gtk.Template.Child()
+    loading_details = Gtk.Template.Child()
     close_button = Gtk.Template.Child()
     stop_button = Gtk.Template.Child()
     help_button = Gtk.Template.Child()
     about_button = Gtk.Template.Child()
     show_hidden_button = Gtk.Template.Child()
     a_to_z_button = Gtk.Template.Child()
+    last_modified_button = Gtk.Template.Child()
     go_top_button = Gtk.Template.Child()
     about_back_button = Gtk.Template.Child()
     properties_back_button = Gtk.Template.Child()
@@ -80,7 +91,9 @@ class PortfolioWindow(Handy.ApplicationWindow):
     selection_tools = Gtk.Template.Child()
     navigation_tools = Gtk.Template.Child()
     places_box = Gtk.Template.Child()
+    places_deck = Gtk.Template.Child()
     places_inner_box = Gtk.Template.Child()
+    places_popup_box = Gtk.Template.Child()
     content_stack = Gtk.Template.Child()
     loading_box = Gtk.Template.Child()
     content_box = Gtk.Template.Child()
@@ -90,6 +103,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
     close_tools = Gtk.Template.Child()
     stop_box = Gtk.Template.Child()
     stop_tools = Gtk.Template.Child()
+    trash_tools = Gtk.Template.Child()
     about_deck = Gtk.Template.Child()
     content_deck = Gtk.Template.Child()
     headerbar = Gtk.Template.Child()
@@ -108,6 +122,12 @@ class PortfolioWindow(Handy.ApplicationWindow):
     property_created = Gtk.Template.Child()
     property_modified = Gtk.Template.Child()
     property_accessed = Gtk.Template.Child()
+    passphrase_header = Gtk.Template.Child()
+    passphrase_box = Gtk.Template.Child()
+    passphrase_entry = Gtk.Template.Child()
+    passphrase_label = Gtk.Template.Child()
+    passphrase_spinner = Gtk.Template.Child()
+    passphrase_back_button = Gtk.Template.Child()
 
     ICON_COLUMN = 0
     NAME_COLUMN = 1
@@ -118,11 +138,13 @@ class PortfolioWindow(Handy.ApplicationWindow):
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
         self._setup()
+        self._setup_settings()
 
     def _setup(self):
         Handy.init()
 
         self._popup = None
+        self._places_popup = None
         self._worker = None
         self._busy = False
         self._editing = False
@@ -130,6 +152,8 @@ class PortfolioWindow(Handy.ApplicationWindow):
         self._to_cut = []
         self._to_select = None
         self._to_select_row = None
+        self._to_go_to = None
+        self._to_go_to_row = None
         self._last_clicked = None
         self._last_vscroll_value = None
         self._dont_activate = False
@@ -139,6 +163,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
         self._index = -1
         self._search_delay_handler_id = 0
         self._load_delay_handler_id = 0
+        self._encrypted = None
 
         self.gesture = Gtk.GestureLongPress.new(self.treeview)
         self.gesture.connect("pressed", self._on_long_pressed)
@@ -162,6 +187,8 @@ class PortfolioWindow(Handy.ApplicationWindow):
         self.cut.connect("clicked", self._on_cut_clicked)
         self.copy.connect("clicked", self._on_copy_clicked)
         self.paste.connect("clicked", self._on_paste_clicked)
+        self.restore_trash.connect("clicked", self._on_restore_trash_clicked)
+        self.delete_trash.connect("clicked", self._on_delete_trash_clicked)
         self.select_all.connect("clicked", self._on_select_all)
         self.select_none.connect("clicked", self._on_select_none)
         self.new_folder.connect("clicked", self._on_new_folder)
@@ -174,6 +201,8 @@ class PortfolioWindow(Handy.ApplicationWindow):
         self.stop_button.connect("clicked", self._on_stop_clicked)
         self.about_back_button.connect("clicked", self._on_about_back_clicked)
         self.properties_back_button.connect("clicked", self._on_properties_back_clicked)
+        self.passphrase_back_button.connect("clicked", self._on_passphrase_back_clicked)
+        self.passphrase_entry.connect("activate", self._on_passphrase_activate)
 
         # XXX no model for options yet so this...
         self.menu_button.connect("clicked", self._on_menu_button_clicked)
@@ -188,7 +217,10 @@ class PortfolioWindow(Handy.ApplicationWindow):
 
         places = PortfolioPlaces()
         places.connect("updated", self._on_places_updated)
+        places.connect("removing", self._on_places_removing)
         places.connect("removed", self._on_places_removed)
+        places.connect("failed", self._on_places_failed)
+        places.connect("unlock", self._on_places_unlock)
         self.places_inner_box.add(places)
 
         self._properties = PortfolioPropertiesWorker()
@@ -238,6 +270,16 @@ class PortfolioWindow(Handy.ApplicationWindow):
         self.content_deck.connect("notify::visible-child", self._on_content_folded)
         self.connect("destroy", self._on_shutdown)
 
+    def _setup_settings(self):
+        self._settings = PortfolioSettings()
+
+        self.show_hidden_button.props.active = self._settings.show_hidden
+
+        if self._settings.sort_order == PortfolioSettings.ALPHABETICAL_ORDER:
+            self.a_to_z_button.props.active = True
+        else:
+            self.last_modified_button.props.active = True
+
     def _filter(self, model, row, data=None):
         path = model[row][self.PATH_COLUMN]
         text = self.search_entry.get_text()
@@ -246,8 +288,8 @@ class PortfolioWindow(Handy.ApplicationWindow):
         return text.lower() in os.path.basename(path).lower()
 
     def _sort_by_last_modified(self, path1, path2):
-        st_mtime1 = os.path.getmtime(path1)
-        st_mtime2 = os.path.getmtime(path2)
+        st_mtime1 = utils.get_file_mtime(path1)
+        st_mtime2 = utils.get_file_mtime(path2)
 
         if st_mtime1 < st_mtime2:
             return 1
@@ -271,8 +313,8 @@ class PortfolioWindow(Handy.ApplicationWindow):
         path1 = model[row1][self.PATH_COLUMN]
         path2 = model[row2][self.PATH_COLUMN]
 
-        row1_is_dir = os.path.isdir(path1)
-        row2_is_dir = os.path.isdir(path2)
+        row1_is_dir = utils.is_file_dir(path1)
+        row2_is_dir = utils.is_file_dir(path2)
 
         if row1_is_dir and not row2_is_dir:
             return -1
@@ -347,9 +389,12 @@ class PortfolioWindow(Handy.ApplicationWindow):
         if self._worker is not None:
             self._worker.stop()
 
-        self._worker = PortfolioLoadWorker(
-            directory, self.show_hidden_button.props.active
-        )
+        if default_trash.is_trash(directory):
+            loader_class = PortfolioLoadTrashWorker
+        else:
+            loader_class = PortfolioLoadWorker
+
+        self._worker = loader_class(directory, self.show_hidden_button.props.active)
         self._worker.connect("started", self._on_load_started)
         self._worker.connect("updated", self._on_load_updated)
         self._worker.connect("finished", self._on_load_finished)
@@ -396,6 +441,14 @@ class PortfolioWindow(Handy.ApplicationWindow):
     def _get_path(self, model, treepath):
         return model[model.get_iter(treepath)][self.PATH_COLUMN]
 
+    def _go_to(self, row):
+        result, row = self.filtered.convert_child_iter_to_iter(row)
+        result, row = self.sorted.convert_child_iter_to_iter(row)
+
+        treepath = self.sorted.get_path(row)
+
+        self.treeview.scroll_to_cell(treepath, None, False, 0, 0)
+
     def _go_to_selection(self):
         model, treepaths = self.selection.get_selected_rows()
         treepath = treepaths[-1]
@@ -409,14 +462,14 @@ class PortfolioWindow(Handy.ApplicationWindow):
             self.treeview.scroll_to_cell(0, None, True, 0, 0)
 
     def _go_back_to_homepage(self):
-        self.content_deck.set_visible_child(self.places_box)
+        self.content_deck.set_visible_child(self.places_deck)
 
     def _move(self, path, navigating=False):
         self._clean_popups()
 
         if path is None:
             return
-        elif os.path.isdir(path):
+        elif default_trash.is_trash(path) or utils.is_file_dir(path):
             self._update_history(path, navigating)
             self._populate(path)
         else:
@@ -446,17 +499,25 @@ class PortfolioWindow(Handy.ApplicationWindow):
     def _switch_to_selection_mode(self):
         self.selection.set_mode(Gtk.SelectionMode.MULTIPLE)
 
-    def _notify(self, description, on_confirm, on_cancel, autoclose, data):
+    def _notify(self, description, on_confirm, on_cancel, on_trash, autoclose, data):
         self._clean_popups()
 
         self._popup = PortfolioPopup(
-            description, on_confirm, on_cancel, autoclose, data
+            description, on_confirm, on_cancel, on_trash, autoclose, data
         )
         self.popup_box.add(self._popup)
         self._popup.props.reveal_child = True
 
+    def _places_notify(self, description):
+        if self._places_popup is not None:
+            self._places_popup.destroy()
+
+        self._places_popup = PortfolioPopup(description, None, None, None, False, None)
+        self.places_popup_box.add(self._places_popup)
+        self._places_popup.props.reveal_child = True
+
     def _find_icon(self, path):
-        if os.path.isdir(path):
+        if utils.is_file_dir(path):
             return "folder-symbolic"
         else:
             return "text-x-generic-symbolic"
@@ -473,12 +534,23 @@ class PortfolioWindow(Handy.ApplicationWindow):
 
     def _clean_progress(self):
         self.loading_description.set_text("")
+        self.loading_details.set_text("")
+        self.loading_details.props.visible = False
 
     def _clean_loading_delay(self):
         if self._load_delay_handler_id != 0:
             GLib.Source.remove(self._load_delay_handler_id)
             self._load_delay_handler_id = 0
 
+    def _clean_passphrase(self):
+        self._encrypted = None
+        self.passphrase_entry.set_text("")
+        self.passphrase_label.set_text("")
+        self.passphrase_entry.props.sensitive = True
+        self.passphrase_label.props.visible = True
+        self.passphrase_spinner.props.visible = False
+        self.passphrase_spinner.props.active = False
+
     def _update_mode(self):
         count = self.selection.count_selected_rows()
         if count == 0:
@@ -496,6 +568,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
         self._update_content_stack()
         self._update_navigation()
         self._update_navigation_tools()
+        self._update_trash_tools()
         self._update_selection()
         self._update_selection_tools()
         self._update_action_stack()
@@ -547,6 +620,11 @@ class PortfolioWindow(Handy.ApplicationWindow):
         self.action_stack.set_visible_child(child)
 
     def _update_tools_stack(self):
+        directory = self._history[self._index]
+        if default_trash.in_trash(directory):
+            self.tools_stack.set_visible_child(self.trash_tools)
+            return
+
         count = self.selection.count_selected_rows()
         selected = count >= 1
         child = self.selection_tools if selected else self.navigation_tools
@@ -570,6 +648,12 @@ class PortfolioWindow(Handy.ApplicationWindow):
         self.paste.props.sensitive = not selected and to_paste and not self._busy
         self.new_folder.props.sensitive = not selected and not self._busy
 
+    def _update_trash_tools(self):
+        selected = self.selection.count_selected_rows() >= 1
+        is_trash = default_trash.is_trash(self._history[self._index])
+        self.restore_trash.props.sensitive = selected and is_trash
+        self.delete_trash.props.sensitive = selected and is_trash
+
     def _update_rename(self):
         count = self.selection.count_selected_rows()
         sensitive = count == 1 and not self._editing and not self._busy
@@ -582,7 +666,12 @@ class PortfolioWindow(Handy.ApplicationWindow):
 
     def _update_directory_title(self):
         directory = self._history[self._index]
-        name = os.path.basename(directory)
+
+        if default_trash.is_trash(directory):
+            name = PortfolioPlaces.XDG_TRASH_NAME
+        else:
+            name = os.path.basename(directory)
+
         self.headerbar.set_title(name)
 
     def _update_filter(self):
@@ -610,6 +699,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
 
         self.loading_label.set_text(_("Opening"))
         self.loading_bar.set_fraction(0.0)
+        self.loading_details.props.visible = False
         self.content_stack.set_visible_child(self.loading_box)
 
         self._update_all()
@@ -649,6 +739,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
     def _on_load_started_delayed(self):
         self.loading_label.set_text(_("Loading"))
         self.loading_bar.set_fraction(0.0)
+        self.loading_details.props.visible = False
         self.content_stack.set_visible_child(self.loading_box)
 
         self._update_all()
@@ -664,6 +755,9 @@ class PortfolioWindow(Handy.ApplicationWindow):
             if self._to_select == path:
                 self._to_select_row = row
 
+            if self._to_go_to == path:
+                self._to_go_to_row = row
+
         self.loading_bar.set_fraction((index + 1) / total)
 
     def _on_load_finished(self, worker, directory):
@@ -678,6 +772,10 @@ class PortfolioWindow(Handy.ApplicationWindow):
             self._select_and_go(self._to_select_row)
             self._to_select_row = None
             self._to_select = None
+        elif self._to_go_to_row is not None:
+            self._go_to(self._to_go_to_row)
+            self._to_go_to_row = None
+            self._to_go_to = None
         else:
             self._go_to_top()
 
@@ -734,6 +832,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
         if self._index == 0:
             self._go_back_to_homepage()
         else:
+            self._to_go_to = self._history[self._index]
             self._index -= 1
             self._move(self._history[self._index], True)
 
@@ -788,7 +887,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
 
         try:
             # respect empty folders
-            if os.path.exists(new_path):
+            if os.path.lexists(new_path):
                 raise FileExistsError(_("%s already exists") % new_path)
 
             os.rename(old_path, new_path)
@@ -806,6 +905,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
                 _("%s already exists") % new_name,
                 None,
                 self._on_popup_closed,
+                None,
                 True,
                 None,
             )
@@ -829,6 +929,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
     def _on_delete_clicked(self, button):
         selection = self._get_selection()
         count = len(selection)
+        directory = self._history[self._index]
 
         if count == 1:
             name = os.path.basename(selection[0][0])
@@ -841,6 +942,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
             description,
             self._on_delete_confirmed,
             self._on_popup_closed,
+            self._on_trash_instead if default_trash.has_trash(directory) else None,
             False,
             selection,
         )
@@ -858,7 +960,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
         else:
             description = _("%d files will be moved") % count
 
-        self._notify(description, None, None, True, None)
+        self._notify(description, None, None, None, True, None)
 
         self._unselect_all()
         self._update_mode()
@@ -876,7 +978,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
         else:
             description = _("%d files will be copied") % count
 
-        self._notify(description, None, None, True, None)
+        self._notify(description, None, None, None, True, None)
 
         self._unselect_all()
         self._update_mode()
@@ -889,7 +991,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
         should_warn = any(
             [
                 os.path.dirname(path) != directory
-                and os.path.exists(os.path.join(directory, os.path.basename(path)))
+                and os.path.lexists(os.path.join(directory, os.path.basename(path)))
                 for path, ref in to_paste
             ]
         )
@@ -902,6 +1004,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
             _("Files will be overwritten, proceed?"),
             self._on_paste_confirmed,
             self._on_popup_closed,
+            None,
             False,
             (to_paste, Worker),
         )
@@ -916,6 +1019,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
 
         self.loading_label.set_text(_("Pasting"))
         self.loading_bar.set_fraction(0.0)
+        self.loading_details.props.visible = True
         self.content_stack.set_visible_child(self.loading_box)
 
         self._update_all()
@@ -923,18 +1027,25 @@ class PortfolioWindow(Handy.ApplicationWindow):
         self.action_stack.set_visible_child(self.stop_box)
         self.tools_stack.set_visible_child(self.stop_tools)
 
-    def _on_paste_pre_updated(self, worker, path):
-        name = os.path.basename(path)
-        self.loading_description.set_text(name)
+    def _on_paste_pre_updated(self, worker, path, overwritten):
+        if overwritten:
+            return
 
-    def _on_paste_updated(self, worker, path, overwritten, index, total):
         # XXX this approach won't allow me to put stat info in liststore
-        if not overwritten:
-            icon = self._find_icon(path)
-            name = os.path.basename(path)
-            self.liststore.append([icon, name, path])
+        name = os.path.basename(path)
+        icon = self._find_icon(path)
+        self.liststore.append([icon, name, path])
 
-        self.loading_bar.set_fraction((index + 1) / total)
+    def _on_paste_updated(self, worker, path, index, total, current_bytes, total_bytes):
+        description = os.path.basename(path)
+        self.loading_description.set_text(description)
+        self.loading_bar.set_fraction(index / total)
+
+        human_current_bytes = utils.get_size_for_humans(current_bytes)
+        human_total_bytes = utils.get_size_for_humans(total_bytes)
+        self.loading_details.set_text(
+            _("%s of %s") % (human_current_bytes, human_total_bytes)
+        )
 
     def _on_paste_finished(self, worker, total):
         self._paste_finish()
@@ -958,6 +1069,26 @@ class PortfolioWindow(Handy.ApplicationWindow):
         # XXX nuclear fix for when parent directorty doesn't get to be updated
         self._refresh()
 
+    def _on_trash_instead(self, button, popup, selection):
+        self._clean_popups()
+
+        # clean history entries from deleted paths
+        directory = self._history[self._index]
+        self._history = [
+            path
+            for path in self._history
+            if not path.startswith(directory) or path == directory
+        ]
+
+        self._worker = PortfolioSendTrashWorker(selection)
+        self._worker.connect("started", self._on_delete_started)
+        self._worker.connect("pre-update", self._on_delete_pre_updated)
+        self._worker.connect("updated", self._on_delete_updated)
+        self._worker.connect("finished", self._on_delete_finished)
+        self._worker.connect("failed", self._on_delete_failed)
+        self._worker.connect("stopped", self._on_delete_stopped)
+        self._worker.start()
+
     def _on_delete_confirmed(self, button, popup, selection):
         self._clean_popups()
 
@@ -983,6 +1114,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
 
         self.loading_label.set_text(_("Deleting"))
         self.loading_bar.set_fraction(0.0)
+        self.loading_details.props.visible = False
         self.content_stack.set_visible_child(self.loading_box)
 
         self._update_all()
@@ -1054,6 +1186,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
                 _("No permissions on this directory"),
                 None,
                 self._on_popup_closed,
+                None,
                 True,
                 None,
             )
@@ -1065,6 +1198,151 @@ class PortfolioWindow(Handy.ApplicationWindow):
         row = self.liststore.append([icon, folder_name, path])
         self._select_and_go(row, edit=True)
 
+    def _on_restore_trash_clicked(self, button):
+        selection = self._get_selection()
+        paths = [default_trash.get_orig_path(path) for path, ref in selection]
+
+        overwrites = any([os.path.lexists(path) for path in paths if path])
+        duplicates = len(set(paths)) != len(paths)
+
+        if not overwrites and not duplicates:
+            self._on_restore_trash_confirmed(None, None, selection)
+            return
+
+        self._notify(
+            _("Files will be overwritten, proceed?"),
+            self._on_restore_trash_confirmed,
+            self._on_popup_closed,
+            None,
+            False,
+            selection,
+        )
+
+    def _on_restore_trash_confirmed(self, button, popup, selection):
+        self._clean_popups()
+
+        self._worker = PortfolioRestoreTrashWorker(selection)
+        self._worker.connect("started", self._on_restore_trash_started)
+        self._worker.connect("pre-update", self._on_restore_trash_pre_updated)
+        self._worker.connect("updated", self._on_restore_trash_updated)
+        self._worker.connect("finished", self._on_restore_trash_finished)
+        self._worker.connect("failed", self._on_restore_trash_failed)
+        self._worker.connect("stopped", self._on_restore_trash_stopped)
+        self._worker.start()
+
+    def _on_restore_trash_started(self, worker):
+        self._busy = True
+
+        self.loading_label.set_text(_("Restoring"))
+        self.loading_bar.set_fraction(0.0)
+        self.loading_details.props.visible = False
+        self.content_stack.set_visible_child(self.loading_box)
+
+        self._update_all()
+
+        self.action_stack.set_visible_child(self.stop_box)
+        self.tools_stack.set_visible_child(self.stop_tools)
+
+    def _on_restore_trash_pre_updated(self, worker, path):
+        name = os.path.basename(path)
+        self.loading_description.set_text(name)
+
+    def _on_restore_trash_updated(self, worker, path, ref, index, total):
+        self._remove_row(ref)
+        self.loading_bar.set_fraction((index + 1) / total)
+
+    def _on_restore_trash_finished(self, worker, total):
+        self._busy = False
+        self._clean_workers()
+        self._clean_progress()
+
+        self._unselect_all()
+
+        self._update_all()
+        self._update_mode()
+
+    def _on_restore_trash_failed(self, worker, path):
+        self._busy = False
+        self._clean_workers()
+        self._clean_progress()
+
+        self.loading_description.set_text(_("Could not restore %s") % path)
+
+        self.action_stack.set_visible_child(self.close_box)
+        self.tools_stack.set_visible_child(self.close_tools)
+
+    def _on_restore_trash_stopped(self, worker):
+        self._delete_finish()
+
+    def _on_delete_trash_clicked(self, button):
+        selection = self._get_selection()
+        count = len(selection)
+
+        if count == 1:
+            name = os.path.basename(selection[0][0])
+        else:
+            name = _("these %d files") % count
+
+        description = _("Delete %s permanently?") % name
+
+        self._notify(
+            description,
+            self._on_delete_trash_confirmed,
+            self._on_popup_closed,
+            None,
+            False,
+            selection,
+        )
+
+    def _on_delete_trash_confirmed(self, button, popup, selection):
+        self._clean_popups()
+
+        self._worker = PortfolioDeleteTrashWorker(selection)
+        self._worker.connect("started", self._on_delete_trash_started)
+        self._worker.connect("pre-update", self._on_delete_trash_pre_updated)
+        self._worker.connect("updated", self._on_delete_trash_updated)
+        self._worker.connect("finished", self._on_delete_trash_finished)
+        self._worker.connect("failed", self._on_delete_trash_failed)
+        self._worker.connect("stopped", self._on_delete_trash_stopped)
+        self._worker.start()
+
+    def _on_delete_trash_started(self, worker):
+        self._busy = True
+
+        self.loading_label.set_text(_("Deleting"))
+        self.loading_bar.set_fraction(0.0)
+        self.loading_details.props.visible = False
+        self.content_stack.set_visible_child(self.loading_box)
+
+        self._update_all()
+
+        self.action_stack.set_visible_child(self.stop_box)
+        self.tools_stack.set_visible_child(self.stop_tools)
+
+    def _on_delete_trash_pre_updated(self, worker, path):
+        name = os.path.basename(path)
+        self.loading_description.set_text(name)
+
+    def _on_delete_trash_updated(self, worker, path, ref, index, total):
+        self._remove_row(ref)
+        self.loading_bar.set_fraction((index + 1) / total)
+
+    def _on_delete_trash_finished(self, worker, total):
+        self._delete_finish()
+
+    def _on_delete_trash_failed(self, worker, path):
+        self._busy = False
+        self._clean_workers()
+        self._clean_progress()
+
+        self.loading_description.set_text(_("Could not delete %s") % path)
+
+        self.action_stack.set_visible_child(self.close_box)
+        self.tools_stack.set_visible_child(self.close_tools)
+
+    def _on_delete_trash_stopped(self, worker):
+        self._delete_finish()
+
     def _on_row_activated(self, treeview, treepath, treecolumn, data=None):
         if self._dont_activate is True:
             self._dont_activate = False
@@ -1077,14 +1355,56 @@ class PortfolioWindow(Handy.ApplicationWindow):
         self._reset_to_path(path)
         self.content_deck.set_visible_child(self.files_box)
 
-    def _on_places_removed(self, button, path):
+    def _on_places_removing(self, button, path):
+        self._places_notify(_("Removing device, please wait"))
+
+    def _on_places_removed(self, button, path, safely):
+        if safely is True:
+            self._places_notify(_("Device can be removed"))
+
         if self._index == -1:
             return
+
         directory = self._history[self._index]
-        if not directory.startswith(path):
+        if not path or not directory.startswith(path):
             return
+
         self._go_back_to_homepage()
 
+    def _on_places_failed(self, button, path):
+        self._places_notify(_("Device is busy, can't be removed"))
+
+    def _on_places_unlock(self, button, encrypted):
+        self._clean_passphrase()
+        self._encrypted = encrypted
+        self.passphrase_header.props.title = encrypted.get_friendly_label()
+        self.places_deck.set_visible_child(self.passphrase_box)
+        self.passphrase_entry.grab_focus()
+
+    def _on_passphrase_back_clicked(self, button):
+        self._clean_passphrase()
+        self.places_deck.set_visible_child(self.places_box)
+
+    def _on_passphrase_activate(self, button):
+        self.passphrase_entry.props.sensitive = False
+        self.passphrase_label.props.visible = False
+        self.passphrase_spinner.props.visible = True
+        self.passphrase_spinner.props.active = True
+
+        passphrase = self.passphrase_entry.get_text()
+        self._encrypted.unlock(passphrase, self._on_places_unlock_finished)
+
+    def _on_places_unlock_finished(self, encrypted, success):
+        self._clean_passphrase()
+
+        if success is True:
+            self._on_passphrase_back_clicked(None)
+            return
+
+        self._encrypted = encrypted
+        self.passphrase_entry.grab_focus()
+        self.passphrase_label.set_text(_("Sorry, that didn't work"))
+
     def _on_help_clicked(self, button):
         Gio.AppInfo.launch_default_for_uri("https://github.com/tchx84/Portfolio", None)
 
@@ -1116,9 +1436,15 @@ class PortfolioWindow(Handy.ApplicationWindow):
         self._update_mode()
 
     def _on_hidden_toggled(self, button):
+        self._settings.show_hidden = self.show_hidden_button.props.active
         self._refresh()
 
     def _on_sort_toggled(self, button):
+        if self.a_to_z_button.props.active:
+            self._settings.sort_order = PortfolioSettings.ALPHABETICAL_ORDER
+        else:
+            self._settings.sort_order = PortfolioSettings.MODIFIED_TIME_ORDER
+
         self._refresh()
 
     def _on_menu_button_clicked(self, button):
@@ -1127,8 +1453,9 @@ class PortfolioWindow(Handy.ApplicationWindow):
 
     def _on_content_folded(self, deck, data=None):
         child = self.content_deck.get_visible_child()
-        if child == self.places_box and self._worker is not None:
+        if child == self.places_deck and self._worker is not None:
             self._worker.stop()
+            self._clean_workers()
         elif child == self.files_box:
             self._properties.stop()
 
@@ -1138,8 +1465,7 @@ class PortfolioWindow(Handy.ApplicationWindow):
         self._properties.stop()
 
     def open(self, path=PortfolioPlaces.PORTFOLIO_HOME_DIR, force_page_switch=False):
-        # XXX so cheap !
-        path = path.replace("file://", "")
+        path = utils.get_uri_path(path)
 
         # make sure it exists though !
         if not os.path.exists(path):
