@@ -16,22 +16,34 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 import os
+import sys
+import time
 import shutil
 import locale
 import datetime
 import threading
 
-from locale import gettext as _
 from gi.repository import Gio, GObject, GLib
 
 from . import utils
 from . import logger
+from .cache import default_cache
+from .translation import gettext as _
+from .trash import default_trash
 
 
 class WorkerStoppedException(Exception):
     pass
 
 
+class CachedWorker(object):
+    def __init__(self):
+        default_cache.activate()
+
+    def __del__(self):
+        default_cache.deactivate()
+
+
 class PortfolioWorker(threading.Thread, GObject.GObject):
     __gtype_name__ = "PortfolioWorker"
 
@@ -43,17 +55,6 @@ class PortfolioWorker(threading.Thread, GObject.GObject):
     def _progress(self, current, total):
         logger.debug(current, total)
 
-    def _copy(self, source, destination):
-        source = Gio.File.new_for_path(source)
-        destination = Gio.File.new_for_path(destination)
-
-        try:
-            source.copy(destination, Gio.FileCopyFlags.OVERWRITE, self._cancellable)
-        except GLib.Error as e:
-            if e.matches(Gio.io_error_quark(), Gio.IOErrorEnum.CANCELLED):
-                raise WorkerStoppedException()
-            raise
-
     def _stop_check(self):
         if self._cancellable.is_cancelled():
             raise WorkerStoppedException()
@@ -71,8 +72,8 @@ class PortfolioCopyWorker(PortfolioWorker):
 
     __gsignals__ = {
         "started": (GObject.SignalFlags.RUN_LAST, None, (int,)),
-        "pre-update": (GObject.SignalFlags.RUN_LAST, None, (str,)),
-        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, bool, int, int)),
+        "pre-update": (GObject.SignalFlags.RUN_LAST, None, (str, bool)),
+        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, int, int, float, float)),
         "finished": (GObject.SignalFlags.RUN_LAST, None, (int,)),
         "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
         "stopped": (GObject.SignalFlags.RUN_LAST, None, ()),
@@ -83,36 +84,99 @@ class PortfolioCopyWorker(PortfolioWorker):
         self._selection = selection
         self._directory = directory
 
+    def _do_copy(self, source, destination, callback):
+        current_bytes = 0
+        total_bytes = os.stat(source.name).st_size
+        arch_capped = sys.maxsize < 2**32
+
+        # start with an arbitrary block size
+        block_min_bytes = 2**23
+        block_bytes = block_min_bytes
+        last_sync = time.monotonic()
+
+        outfd = destination.fileno()
+        infd = source.fileno()
+
+        while True:
+            self._stop_check()
+
+            # assume kernel >= 2.6.33
+            sent_bytes = os.sendfile(outfd, infd, current_bytes, block_bytes)
+            if sent_bytes == 0:
+                break
+
+            # fsync frequency is indirectly controlled by the size of the block
+            os.fsync(outfd)
+
+            # now dynamically adjust the block size so that it
+            # will keep fsync frequency between 750 and 1250ms
+            now = time.monotonic()
+            if now - last_sync > 1.25:
+                block_bytes //= 2
+            elif now - last_sync < 0.75:
+                block_bytes *= 2
+
+            # on 32 bit arch block size must be capped
+            if arch_capped:
+                block_bytes = min(block_bytes, 2**30)
+
+            # some lower and upper bounds
+            block_bytes = max(block_bytes, block_min_bytes)
+            block_bytes = min(block_bytes, total_bytes)
+
+            last_sync = now
+            current_bytes += sent_bytes
+            callback(current_bytes, total_bytes)
+
+        logger.debug(f"sendfile block size converged to {block_bytes}")
+
+    def _copy(self, source_path, destination_path):
+        if os.path.islink(source_path):
+            os.symlink(os.readlink(source_path), destination_path)
+            return
+
+        def callback(current_bytes, total_bytes):
+            self.emit(
+                "updated",
+                destination_path,
+                self._count,
+                self._total,
+                current_bytes,
+                total_bytes,
+            )
+
+        with open(source_path, "rb") as source:
+            with open(destination_path, "wb") as destination:
+                self._do_copy(source, destination, callback)
+
+        shutil.copymode(source_path, destination_path)
+
+        self._count += 1
+
     def run(self):
-        count = 0
-        total = sum([utils.count(path) for path, ref in self._selection])
-        self.emit("started", total)
+        self._count = 0
+        self._total = sum([utils.count(path) for path, ref in self._selection])
+
+        self.emit("started", self._total)
 
         for path, ref in self._selection:
             name = os.path.basename(path)
             destination = os.path.join(self._directory, name)
-            overwritten = os.path.exists(destination)
+            overwritten = os.path.lexists(destination)
 
             if path == destination:
                 name = utils.find_new_name(self._directory, name)
                 destination = os.path.join(self._directory, name)
                 overwritten = False
 
-            def _callback(_path, _destination):
-                nonlocal count, total
-                self._stop_check()
-                self.emit("pre-update", _destination)
-                self._copy(_path, _destination)
-                self.emit("updated", _destination, True, count, total)
-                count += 1
-
             try:
                 self._stop_check()
-                self.emit("pre-update", destination)
+                self.emit("pre-update", destination, overwritten)
+
                 if os.path.isdir(path):
                     if overwritten and os.path.isdir(path):
                         shutil.rmtree(destination)
-                    shutil.copytree(path, destination, copy_function=_callback)
+                    shutil.copytree(path, destination, copy_function=self._copy)
                 else:
                     self._copy(path, destination)
             except WorkerStoppedException:
@@ -122,42 +186,37 @@ class PortfolioCopyWorker(PortfolioWorker):
                 logger.debug(e)
                 self.emit("failed", destination)
                 return
-            else:
-                self.emit("updated", destination, overwritten, count, total)
-                count += 1
+            finally:
+                utils.sync_folder(os.path.dirname(destination))
 
-        self.emit("finished", total)
+        self.emit("finished", self._total)
 
 
 class PortfolioCutWorker(PortfolioCopyWorker):
     __gtype_name__ = "PortfolioCutWorker"
 
     def run(self):
-        count = 0
-        total = sum([utils.count(path) for path, ref in self._selection])
-        self.emit("started", total)
+        self._count = 0
+        self._total = sum([utils.count(path) for path, ref in self._selection])
+
+        self.emit("started", self._total)
 
         for path, ref in self._selection:
             name = os.path.basename(path)
             destination = os.path.join(self._directory, name)
-            overwritten = os.path.exists(destination)
-
-            def _callback(_path, _destination):
-                nonlocal count, total
-                self._stop_check()
-                self.emit("pre-update", _destination)
-                self._copy(_path, _destination)
-                self.emit("updated", _destination, True, count, total)
-                count += 1
+            overwritten = os.path.lexists(destination)
 
             try:
                 self._stop_check()
-                self.emit("pre-update", destination)
+                self.emit("pre-update", destination, overwritten)
+
                 if destination == path:
                     continue
                 if overwritten and os.path.isdir(path):
                     shutil.rmtree(destination)
-                shutil.move(path, destination, copy_function=_callback)
+                if overwritten and os.path.islink(destination):
+                    os.unlink(destination)
+                shutil.move(path, destination, copy_function=self._copy)
             except WorkerStoppedException:
                 self.emit("stopped")
                 return
@@ -165,14 +224,13 @@ class PortfolioCutWorker(PortfolioCopyWorker):
                 logger.debug(e)
                 self.emit("failed", path)
                 return
-            else:
-                self.emit("updated", destination, overwritten, count, total)
-                count += 1
+            finally:
+                utils.sync_folder(os.path.dirname(destination))
 
-        self.emit("finished", total)
+        self.emit("finished", self._total)
 
 
-class PortfolioDeleteWorker(GObject.GObject):
+class PortfolioDeleteWorker(GObject.GObject, CachedWorker):
     __gtype_name__ = "PortfolioDeleteWorker"
 
     __gsignals__ = {
@@ -186,6 +244,7 @@ class PortfolioDeleteWorker(GObject.GObject):
 
     def __init__(self, selection):
         super().__init__()
+        CachedWorker.__init__(self)
         self._selection = selection
         self._timeout_handler_id = None
 
@@ -241,7 +300,7 @@ class PortfolioDeleteWorker(GObject.GObject):
         self.emit("stopped")
 
 
-class PortfolioLoadWorker(GObject.GObject):
+class PortfolioLoadWorker(GObject.GObject, CachedWorker):
     __gtype_name__ = "PortfolioLoadWorker"
 
     __gsignals__ = {
@@ -255,6 +314,7 @@ class PortfolioLoadWorker(GObject.GObject):
 
     def __init__(self, directory, hidden=False):
         super().__init__()
+        CachedWorker.__init__(self)
         self._directory = directory
         self._hidden = hidden
         self._timeout_handler_id = None
@@ -316,11 +376,14 @@ class PortfolioOpenWorker(GObject.GObject):
         super().__init__()
         self._path = path
         self._timeout_handler_id = None
+        self._is_flatpak = utils.is_flatpak()
 
     def start(self):
         self.emit("started")
+
+        uri = f"file://{self._path}"
         Gio.AppInfo.launch_default_for_uri_async(
-            f"file://{self._path}", None, None, self._on_launch_finished, None
+            uri, None, None, self._on_launch_finished, None
         )
         self._timeout_handler_id = GLib.timeout_add(100, self._on_step)
 
@@ -360,11 +423,11 @@ class PortfolioPropertiesWorker(GObject.GObject):
                     if self._stop:
                         return
                     try:
-                        size += os.path.getsize(os.path.join(directory, filename))
+                        size += os.lstat(os.path.join(directory, filename)).st_size
                     except:
                         pass
 
-            self._worker._size = self._worker._human_size(size)
+            self._worker._size = utils.get_size_for_humans(size)
             GLib.idle_add(self._worker.notify, "size")
 
     def __init__(self):
@@ -382,24 +445,16 @@ class PortfolioPropertiesWorker(GObject.GObject):
         self._inner_worker = self.InnerWorker(self)
 
     def _get_file_size(self):
-        return os.path.getsize(self._path)
+        return os.lstat(self._path).st_size
 
     def _update_size(self):
         if os.path.isdir(self._path):
             self._inner_worker = self.InnerWorker(self)
             self._inner_worker.start()
         else:
-            self._size = self._human_size(self._get_file_size())
+            self._size = utils.get_size_for_humans(self._get_file_size())
             self.notify("size")
 
-    # https://gist.github.com/cbwar/d2dfbc19b140bd599daccbe0fe925597
-    def _human_size(self, num):
-        for unit in ["", "k", "M", "G", "T", "P", "E", "Z"]:
-            if abs(num) < 1024.0:
-                return "%3.1f %sB" % (num, unit)
-            num /= 1024.0
-        return "%.1f%sB" % (num, "Yi")
-
     def _get_human_time(self, timestamp):
         fmt = locale.nl_langinfo(locale.D_T_FMT)
         return datetime.datetime.fromtimestamp(timestamp).strftime(fmt)
@@ -445,7 +500,7 @@ class PortfolioPropertiesWorker(GObject.GObject):
         file = Gio.File.new_for_path(path)
         info = file.query_info(
             Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE,
-            Gio.FileQueryInfoFlags.NONE,
+            Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS,
             None,
         )
 
@@ -453,9 +508,9 @@ class PortfolioPropertiesWorker(GObject.GObject):
         self._location = os.path.dirname(self._path)
         self._type = info.get_content_type()
         self._size = _("Calculating...")
-        self._created = self._get_human_time(os.path.getctime(self._path))
-        self._modified = self._get_human_time(os.path.getmtime(self._path))
-        self._accessed = self._get_human_time(os.path.getatime(self._path))
+        self._created = self._get_human_time(os.lstat(self._path).st_ctime)
+        self._modified = self._get_human_time(os.lstat(self._path).st_mtime)
+        self._accessed = self._get_human_time(os.lstat(self._path).st_atime)
 
         self.notify("name")
         self.notify("location")
@@ -469,3 +524,226 @@ class PortfolioPropertiesWorker(GObject.GObject):
 
     def stop(self):
         self._inner_worker.stop()
+
+
+class PortfolioSendTrashWorker(GObject.GObject, CachedWorker):
+    __gtype_name__ = "PortfolioSendTrashWorker"
+
+    __gsignals__ = {
+        "started": (GObject.SignalFlags.RUN_LAST, None, ()),
+        "pre-update": (GObject.SignalFlags.RUN_LAST, None, (str,)),
+        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, object, int, int)),
+        "finished": (GObject.SignalFlags.RUN_LAST, None, (int,)),
+        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
+        "stopped": (GObject.SignalFlags.RUN_LAST, None, ()),
+    }
+
+    def __init__(self, selection):
+        super().__init__()
+        CachedWorker.__init__(self)
+        self._selection = selection
+        self._is_flatpak = utils.is_flatpak()
+
+    def start(self):
+        self.emit("started")
+
+        self._paths = [path for path, ref in self._selection]
+        self._refs = dict(self._selection)
+
+        self._total = len(self._paths)
+        self._index = 0
+        self._timeout_handler_id = GLib.idle_add(
+            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
+        )
+
+    def step(self):
+        if self._index >= self._total:
+            self.emit("finished", self._total)
+            return
+
+        path = self._paths[self._index]
+
+        try:
+            self.emit("pre-update", path)
+            default_trash.trash(path)
+        except Exception as e:
+            logger.debug(e)
+            self.emit("failed", path)
+            return
+
+        self._index += 1
+        self.emit("updated", path, self._refs.get(path), self._index, self._total)
+        self._timeout_handler_id = GLib.idle_add(
+            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
+        )
+
+    def stop(self):
+        if self._timeout_handler_id is not None:
+            GLib.Source.remove(self._timeout_handler_id)
+            self._timeout_handler_id = None
+        self.emit("stopped")
+
+
+class PortfolioRestoreTrashWorker(GObject.GObject, CachedWorker):
+    __gtype_name__ = "PortfolioRestoreTrashWorker"
+
+    __gsignals__ = {
+        "started": (GObject.SignalFlags.RUN_LAST, None, ()),
+        "pre-update": (GObject.SignalFlags.RUN_LAST, None, (str,)),
+        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, object, int, int)),
+        "finished": (GObject.SignalFlags.RUN_LAST, None, (int,)),
+        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
+        "stopped": (GObject.SignalFlags.RUN_LAST, None, ()),
+    }
+
+    def __init__(self, selection):
+        super().__init__()
+        CachedWorker.__init__(self)
+        self._selection = selection
+
+    def start(self):
+        self.emit("started")
+
+        self._paths = [path for path, ref in self._selection]
+        self._refs = dict(self._selection)
+
+        self._total = len(self._paths)
+        self._index = 0
+        self._timeout_handler_id = GLib.idle_add(
+            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
+        )
+
+    def step(self):
+        if self._index >= self._total:
+            self.emit("finished", self._total)
+            return
+
+        path = self._paths[self._index]
+
+        try:
+            self.emit("pre-update", path)
+            default_trash.restore(path)
+        except Exception as e:
+            logger.debug(e)
+            self.emit("failed", path)
+            return
+
+        self._index += 1
+        self.emit("updated", path, self._refs.get(path), self._index, self._total)
+        self._timeout_handler_id = GLib.idle_add(
+            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
+        )
+
+    def stop(self):
+        if self._timeout_handler_id is not None:
+            GLib.Source.remove(self._timeout_handler_id)
+            self._timeout_handler_id = None
+        self.emit("stopped")
+
+
+class PortfolioDeleteTrashWorker(GObject.GObject):
+    __gtype_name__ = "PortfolioDeleteTrashWorker"
+
+    __gsignals__ = {
+        "started": (GObject.SignalFlags.RUN_LAST, None, ()),
+        "pre-update": (GObject.SignalFlags.RUN_LAST, None, (str,)),
+        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, object, int, int)),
+        "finished": (GObject.SignalFlags.RUN_LAST, None, (int,)),
+        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
+        "stopped": (GObject.SignalFlags.RUN_LAST, None, ()),
+    }
+
+    def __init__(self, selection):
+        super().__init__()
+        self._selection = selection
+
+    def start(self):
+        self.emit("started")
+
+        self._paths = [path for path, ref in self._selection]
+        self._refs = dict(self._selection)
+
+        self._total = len(self._paths)
+        self._index = 0
+        self._timeout_handler_id = GLib.idle_add(
+            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
+        )
+
+    def step(self):
+        if self._index >= self._total:
+            self.emit("finished", self._total)
+            return
+
+        path = self._paths[self._index]
+
+        try:
+            self.emit("pre-update", path)
+            default_trash.remove(path)
+        except Exception as e:
+            logger.debug(e)
+            self.emit("failed", path)
+            return
+
+        self._index += 1
+        self.emit("updated", path, self._refs.get(path), self._index, self._total)
+        self._timeout_handler_id = GLib.idle_add(
+            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
+        )
+
+    def stop(self):
+        if self._timeout_handler_id is not None:
+            GLib.Source.remove(self._timeout_handler_id)
+            self._timeout_handler_id = None
+        self.emit("stopped")
+
+
+class PortfolioLoadTrashWorker(GObject.GObject, CachedWorker):
+    __gtype_name__ = "PortfolioLoadTrashWorker"
+
+    __gsignals__ = {
+        "started": (GObject.SignalFlags.RUN_LAST, None, (str,)),
+        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, object, int, int)),
+        "finished": (GObject.SignalFlags.RUN_LAST, None, (str,)),
+        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
+    }
+
+    def __init__(self, directory=None, hidden=False):
+        super().__init__()
+        CachedWorker.__init__(self)
+        self._timeout_handler_id = None
+
+    def start(self):
+        self.emit("started", "")
+
+        try:
+            self._paths = default_trash.list()
+        except Exception as e:
+            logger.debug(e)
+            self.emit("failed", "")
+            return
+
+        self._total = len(self._paths)
+        self._index = 0
+        self._timeout_handler_id = GLib.idle_add(
+            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
+        )
+
+    def step(self):
+        if self._index >= self._total:
+            self.emit("finished", "")
+            return
+
+        path = self._paths[self._index]
+        name = os.path.basename(path)
+
+        self._index += 1
+        self.emit("updated", "", [(name, path)], self._index, self._total)
+        self._timeout_handler_id = GLib.idle_add(
+            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
+        )
+
+    def stop(self):
+        if self._timeout_handler_id is None:
+            return
+        GLib.Source.remove(self._timeout_handler_id)
+        self._timeout_handler_id = None
